#ifndef _MAIN_MENU_H_
#define _MAIN_MENU_H_

#define ID_FILE						101
#define   ID_RUN_ATARI_PROGRAM		102
#define   ID_RESET					103
#define   ID_REBOOT					104
#define   ID_CONFIGURATION		    105
#define   ID_SAVE_CONFIG            106
#define   ID_BACK                   107
#define   ID_EXIT 					108
#define ID_MANAGE					201
#define   ID_DISK_MANAGEMENT		202
#define   ID_CART_MANAGEMENT		203
#define   ID_TAPE_MANAGEMENT		204
#define   ID_EMULATOR_BIOS			205
#define ID_SYSTEM   				301
#define   ID_NTSC					302
#define   ID_PAL					303
#define   ID_SELECT_MACHINE         304
#define   ID_SOUND                  305
#define   ID_CONTROLLERS			306
#define ID_DISPLAY					401
#define   ID_FULLSCREEN				402
#define   ID_WINDOW_SIZE_UP         403
#define   ID_WINDOW_SIZE_DOWN       404
#define   ID_ATARI_DISPLAY          405
#define   ID_WINDOWS_DISPLAY        406
#define ID_HELP                     901
#define   ID_ABOUT_ATARI800         902
#define   ID_FUNCTION_KEY_HELP      903
#define   ID_HOT_KEY_HELP           904

#endif /* _MENU_H_ */
